/* Copyright (c) 1997-2007
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ident "$Project: polymake $$Id: blending.cc 7861 2008-01-03 23:13:27Z gawrilow $"

#include <polymake.h>
#include <vector>
#include <Array.h>
#include <IncidenceMatrix.h>
#include <Graph.h>

namespace polymake { namespace polytope {

   perl::Object blending(perl::Object p_in1, const int vertex1, perl::Object p_in2, const int vertex2, perl::OptionSet options)
   {
      const int dim=p_in1.give("DIM"),
	       dim2=p_in2.give("DIM");
      if (dim != dim2)
	 throw std::runtime_error("dimension mismatch");

      const IncidenceMatrix<> VIF1=p_in1.give("VERTICES_IN_FACETS"),
	                      VIF2=p_in2.give("VERTICES_IN_FACETS");
      const int n_vertices1=VIF1.cols(), n_facets1=VIF1.rows(),
	        n_vertices2=VIF2.cols();

      if (vertex1 < 0 || vertex1 >= n_vertices1)
	 throw std::runtime_error("first vertex number out of range");
      if (vertex2 < 0 || vertex2 >= n_vertices2)
	 throw std::runtime_error("second vertex number out of range");

      const Graph<> G1=p_in1.give("GRAPH.ADJACENCY"),
	            G2=p_in2.give("GRAPH.ADJACENCY");

      if (G1.degree(vertex1) != dim)
	 throw std::runtime_error("first vertex not simple");
      if (G2.degree(vertex2) != dim)
	 throw std::runtime_error("second vertex not simple");

      Array<int> neighbors2(dim,-1);

      // reorder the neighbor vertices in P2 if required
      Entire<Graph<>::adjacent_node_list>::const_iterator nb=entire(G2.adjacent_nodes(vertex2));
      Array<int> permutation;
      if (options["permutation"] >> permutation) {
	 if (permutation.size() != dim)
	    throw std::runtime_error("wrong permutation size");

	 for (Array<int>::const_iterator p_i=permutation.begin(); !nb.at_end(); ++p_i, ++nb) {
	    if (*p_i < 0 || *p_i >= dim)
	       throw std::runtime_error("permutation element out of range");

	    if (neighbors2[*p_i] >= 0)
	       throw std::runtime_error("permutation not surjective");

	    neighbors2[*p_i] = *nb;
	 }
      } else {
	 std::copy(nb, neighbors2.begin());
      }

      perl::Object p_out("Polytope<Rational>");
      p_out.set_description() << "Blending of " << p_in1.name() << " at vertex " << vertex1 << " and " << p_in2.name() << " at vertex " << vertex2;
      if (permutation.empty())
	 p_out.append_description() << '\n';
      else
	 p_out.append_description() << " permuted with [" << permutation << "]\n";

      // initialize as a block-diagonal matrix
      IncidenceMatrix<> VIF_out=diag(VIF1,VIF2);

      // Facets to be glued are identified via opposite neighbors of vertex{1,2}
      // (that is, not belonging to them).
      // Since both vertices are simple, this mapping is unambigiuos
      Array<int>::const_iterator nb2=neighbors2.begin();
      for (nb=entire(G1.adjacent_nodes(vertex1)); !nb.at_end(); ++nb, ++nb2) {
	 int f1=(VIF1.col(vertex1) - VIF1.col(*nb)).front(),
	     f2=(VIF2.col(vertex2) - VIF2.col(*nb2)).front() + n_facets1;
	 VIF_out.row(f1) += VIF_out.row(f2);
	 VIF_out.row(f2).clear();
      }

      // get rid of the vertices
      VIF_out.col(vertex1).clear();
      VIF_out.col(vertex2+n_vertices1).clear();
      VIF_out.squeeze();

      p_out.take("DIM") << dim;
      p_out.take("N_VERTICES") << n_vertices1+n_vertices2-2;
      p_out.take("VERTICES_IN_FACETS") << VIF_out;

      if (options["relabel"]) {
	 std::vector<std::string> labels1(n_vertices1), labels2(n_vertices2),
	                          labels_out(n_vertices1+n_vertices2-2);
	 read_labels(p_in1, "VERTEX_LABELS", labels1);
	 read_labels(p_in2, "VERTEX_LABELS", labels2);

	 const std::string tick="'";
	 std::copy(entire(concatenate(select(labels1, ~scalar2set(vertex1)),
				      attach_operation(select(labels2, ~scalar2set(vertex2)),
						       constant(tick), operations::add()))),
		   labels_out.begin());

	 p_out.take("VERTEX_LABELS") << labels_out;
      }

      return p_out;
   }

   UserFunction4perl("# Category: Producing a new polyhedron from others\n"
		     "# Compute the blending of two polyhedra at simple vertices.\n"
		     "# This is a slightly less standard construction.\n"
		     "# A vertex is $simple$ if its vertex figure is a simplex.\n"
		     "#\n"
		     "# Moving a vertex $v$ of a bounded polytope to infinity yields an unbounded polyhedron\n"
		     "# with all edges through $v$ becoming mutually parallel rays.  Do this to both\n"
		     "# input polytopes $P$ and $P'$ at simple vertices $v$ and $v'$, respectively.\n"
		     "# Up to an affine transformation one can assume that the orthogonal projections\n"
		     "# of $P$ and $P'$ in direction $v$ and $v'$, respectively, are mutually congruent.\n"
		     "#\n"
		     "# Any bijection $b$ from the set of edges through $v$ to the edges through $v'$\n"
		     "# uniquely defines a way of glueing the unbounded polyhedra to obtain a new bounded\n"
		     "# polytope, the $blending$ with respect to $b$. The bijection is specified as a $permutation$\n"
		     "# of indices $0 1 2$ etc. Default permutation is identity.\n"
		     "#\n"
		     "# The number of vertices of the blending is the sum of the numbers of vertices of the\n"
		     "# input polytopes minus $2$.  The number of facets is the sum of the numbers of facets\n"
		     "# of the input polytopes minus the dimension.\n"
		     "#\n"
		     "# The resulting polytope is described only combinatorially.\n"
		     "#\n"
		     "# @author Kerstin Fritzsche (initial version)\n"
		     "# args: P v P' v' [ permutation => Array<Int> ]\n",
		     &blending, "blending(Polytope $ Polytope $ { permutation => undef, relabel => undef })");
} }
