/* Copyright (c) 1997-2007
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ident "$Project: polymake $$Id: facets_from_incidence.cc 7858 2008-01-02 22:33:38Z gawrilow $"

/** @file facets_from_incidence
 *
 *  Compute the facet description of a polyhedron from its vertices and its incidence matrix.
 */

#include <polymake.h>
#include <IncidenceMatrix.h>
#include <linalg.h>

namespace polymake { namespace polytope {

namespace {

template <typename E, typename Matrix1, typename Vector2> inline
void assign_facet_through_points(const GenericMatrix<Matrix1,E>& V, GenericVector<Vector2,E>& f)
{
   f=null_space(V)[0];
}

template <typename Matrix1, typename Vector2> inline
void assign_facet_through_points(const GenericMatrix<Matrix1,double>& V, GenericVector<Vector2,double>& f)
{
   const Matrix<double> F=null_space(V);
   if (F.rows() == 1) {
      f=F[0];
   } else {
      // least squares
      const int n=V.cols();
      f=inv(( 2*T(V)*V               | ones_vector<double>(n)) /
	    ( ones_vector<double>(n) | 0.0                   ) )[n].slice(0,n);
   }
}

template <typename Scalar, typename IM>
Matrix<Scalar> compute(const Matrix<Scalar>& V, const Matrix<Scalar>& AH,
		       const GenericIncidenceMatrix<IM>& VIF)
{
   Matrix<Scalar> F(VIF.rows(), V.cols());
   typename Rows< Matrix<Scalar> >::iterator Fi=rows(F).begin();

   for (typename Entire< Rows<IM> >::const_iterator vertex_list=entire(rows(VIF));
	!vertex_list.at_end();  ++vertex_list, ++Fi) {
      assign_facet_through_points(V.minor(*vertex_list,All) / AH, Fi->top());
      int outer_vertex=(sequence(0,V.rows()) - *vertex_list).front();
      if (*Fi * V[outer_vertex] < 0) Fi->negate();
   }

   return F;
}

} // end anonymous namespace

template <typename Scalar>
void facets_from_incidence(perl::Object p)
{
   const Matrix<Scalar> V=p.give("VERTICES");
   const IncidenceMatrix<> VIF=p.give("VERTICES_IN_FACETS");

   Matrix<Scalar> AH=null_space(V);
   p.take("AFFINE_HULL") << AH;

   AH.col(0).fill(0);
   p.take("FACETS") << compute(V, AH, VIF);
}

template <typename Scalar>
void vertices_from_incidence(perl::Object p)
{
   const Matrix<Scalar> F=p.give("FACETS"), AH=p.give("AFFINE_HULL");
   const IncidenceMatrix<> VIF=p.give("VERTICES_IN_FACETS");

   p.take("VERTICES") << compute(F, AH, T(VIF));
}

FunctionTemplate4perl("facets_from_incidence<Scalar> (Polytope<Scalar>) : void");
FunctionTemplate4perl("vertices_from_incidence<Scalar> (Polytope<Scalar>) : void");

} }
