/* Copyright (c) 1997-2007
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ident "$Project: polymake $$Id: k-cyclic.cc 7858 2008-01-02 22:33:38Z gawrilow $"

#include <cmath>
#include <polymake.h>
#include <Matrix.h>
#include <Rational.h>
#include <Array.h>

namespace polymake { namespace polytope {

perl::Object k_cyclic(int n, Vector<Rational> s)
{
   const int k=s.size();
   if (k < 1 || k >= n)
      throw std::runtime_error("k_cyclic: 1 <= k < n required");

   perl::Object p("Polytope<Rational>");
   p.set_description() << k << "-cyclic polytope" << endl;

   p.take("AMBIENT_DIM") << 2*k;
   p.take("DIM") << 2*k;
   p.take("N_VERTICES") << n;

   Matrix<Rational> vertices(n,2*k+1);
   Rational* v=concat_rows(vertices).begin();
   s*=2; s/=n;

   for (int i = 0; i < n; ++i) {
      *v++ = 1;
      for (int j = 0; j < k; ++j) {
	 const Rational S=s[j]*i;
	 // prevent silly values for multiples of pi/2
	 if (denominator(S)==1) {
	    *v++ = numerator(S).odd() ? -1 : 1;
	    *v++ = 0;
	 } else if (denominator(S)==2) {
	    *v++ = 0;
	    *v++ = numerator(S).bit(1) ? -1 : 1;
	 } else {
	    const double K=double(S)*M_PI;
	    *v++ = cos(K);
	    *v++ = sin(K);
	 }
      }
   }
   p.take("VERTICES") << vertices;
   p.take("BOUNDED") << true;

   return p;
}

UserFunction4perl("# Category: Producing from scratch\n"
		  "# args: n s_1 .. s_k\n"
		  "# Produce a (rounded) `2*k`-dimensional `k`-cyclic polytope with `n` points.\n"
		  "# Special cases are the bicyclic (`k=2`) and tricyclic (`k=3`) polytopes.\n"
		  "# Only possible in even dimensions.\n"
		  "#\n"
		  "# The parameters `s_i` can be specified as integer, floating-point, or rational numbers.\n"
		  "# The coordinates of the `i`-th point are taken as follows:\n"
		  "#\t `cos(s_1 * 2&pi;i/n)`,\n"
		  "#\t `sin(s_1 * 2&pi;i/n)`,\n"
		  "#\t ...\n"
		  "#\t `cos(s_k * 2&pi;i/n)`,\n"
		  "#\t `sin(s_k * 2&pi;i/n)`\n"
		  "#\n"
		  "# Warning: Some of the `k`-cyclic polytopes are not simplicial.\n"
		  "# Since the components are rounded, this function might output a polytope\n"
		  "# which is not a `k`-cyclic polytope!\n"
		  "#\n"
		  "# More information can be found in the following references:\n"
		  "#\t P. Schuchert: \"Matroid-Polytope und Einbettungen kombinatorischer Mannigfaltigkeiten\",\n"
		  "#\t PhD thesis, TU Darmstadt, 1995.\n"
		  "#\n"
		  "#\t Z. Smilansky: \"Bi-cyclic `4`-polytopes\",\n"
		  "#\t Isr. J. Math. 70, 1990, 82-92\n",
		  &k_cyclic, "k_cyclic($@)");
} }
