/* Copyright (c) 1997-2004
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ident "$Project: polymake $$Id: prism.cc 8000 2008-02-27 14:22:34Z wotzlaw $"

#include <vector>
#include <Rational.h>
#include <Matrix.h>
#include <IncidenceMatrix.h>
#include <perl/Object.h>

namespace polymake { namespace polytope {

Matrix<Rational>
prism_coord(const Matrix<Rational>& V,
	    int& n_vertices, int& n_vertices_out,
	    const Set<int>& rays,
	    const Rational& z, const Rational& z_prime)
{

   if (n_vertices==0) {
      n_vertices=V.rows();
      n_vertices_out=2*n_vertices-rays.size();
   }
   return rays.empty()
          ? Matrix<Rational>( ( V | same_element_vector(z, n_vertices) ) /
			      ( V | same_element_vector(z_prime, n_vertices)) )
          : Matrix<Rational>( ( V | same_element_sparse_vector(~rays, z, n_vertices) ) /
			      ( V.minor(~rays,All) | same_element_vector(z_prime, n_vertices-rays.size()) ) );
}


perl::Object prism(perl::Object p_in, const Rational& z, const Rational& z_prime, perl::OptionSet options)
{

   if(z==z_prime)
	throw std::runtime_error("prism: z and z' must be different");

   int n_vertices=0, n_vertices_out=0;
   Set<int> rays;
   if (!options["noc"])
      p_in.give("FAR_FACE") >> rays;

   perl::Object p_out("Polytope<Rational>");
   p_out.set_description() << "prism over " << p_in.name() << endl;

   if (options["noc"] || p_in.exists("VERTICES_IN_FACETS")) {
      const IncidenceMatrix<> VIF=p_in.give("VERTICES_IN_FACETS");
      n_vertices=VIF.cols();
      n_vertices_out=2*n_vertices-rays.size();

      Set<int> far_facet=accumulate(cols(VIF.minor(All,rays)), operations::mul());
      if (far_facet.size() > 1)	// if size()>1 then FAR_FACE is not a facet
	 far_facet.clear();

      const IncidenceMatrix<> VIF_out=
	 rays.empty()
	 ? IncidenceMatrix<>( (VIF | VIF)				// side facet = original facet && its clone
			      / sequence(0,n_vertices)		// bottom facet = original polytope
			      / sequence(n_vertices,n_vertices) )	// top facet = cloned polytope
	 : IncidenceMatrix<>( (VIF.minor(~far_facet,All) | VIF.minor(~far_facet, ~rays))	// rays aren't cloned
			      / sequence(0,n_vertices)		//   and belong to both (top&bottom) facets
			      / (rays + sequence(n_vertices,n_vertices-rays.size())) );

      p_out.take("N_VERTICES") << n_vertices_out;
      p_out.take("VERTICES_IN_FACETS") << VIF_out;
   }

   if (options["noc"]) {
      if (p_in.exists("DIM")) {
	 const int dim=p_in.give("DIM");
	 p_out.take("DIM") << dim+1;
      }

   } else {
      const Matrix<Rational> V=p_in.give("VERTICES");
      const Matrix<Rational> V_out=prism_coord(V, n_vertices, n_vertices_out, rays, z, z_prime);
      p_out.take("VERTICES") << V_out;
   }

   if (options["relabel"]) {
      std::vector<std::string> labels(n_vertices_out);
      read_labels(p_in, "VERTEX_LABELS", labels);
      const std::string tick="'";
      std::copy(entire(attach_operation(select(labels, sequence(0,n_vertices)-rays),
					constant(tick), operations::add())),
		labels.begin()+n_vertices);
      p_out.take("VERTEX_LABELS") << labels;
   }

   return p_out;
}

UserFunction4perl("# Category:  Producing a new polyhedron from others\n"
		  "# Make a prism over a polytope.\n"
		  "# The prism is the product of the input polytope and the interval [@a z,@a z'].\n"
		  "# The default interval is [$-1,1$].  If @a z != $-1$, then @a z' defaults to @a -z.\n"
		  "# \n"
		  "# If the option @c -noc (no coordinates) is set, then only combinatorial information\n"
		  "# is handled. The input polytope is assumed to be bounded.\n"
		  "# \n"
		  "# The option @a relabel creates an additional section @see VERTEX_LABELS.\n"
		  "# The bottom facet vertices get the labels from the original polytope;\n"
		  "# the labels of their clones in the top facet get a tick (') appended.\n"
		  "# args: [z, [z'] ], [ noc => Bool],[relabel => Bool]\n",
		  &prism, "prism(Polytope; $=-1, $=-$_[1], { noc => undef, relabel => undef})");
} }

