/* Copyright (c) 1997-2008
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ident "$Project: polymake $$Id: projection.cc 8293 2008-08-14 12:28:12Z sherrmann $"

#include <Rational.h>
#include <polymake.h>
#include <Matrix.h>
#include <ListMatrix.h>
#include <PowerSet.h>
#include <linalg.h>
#include <Array.h>


namespace polymake { namespace polytope {

perl::Object projection(perl::Object p_in, const Array<int>& indices, perl::OptionSet options)
{
  const int ambient_dim=p_in.give("AMBIENT_DIM");
  const int dim=p_in.give("DIM");
  const int codim=ambient_dim-dim;
  const Matrix<Rational> affine_hull=p_in.give("AFFINE_HULL");
  if (codim!=affine_hull.rows())
    throw std::runtime_error("p_proj: AFFINE_HULL has wrong number of rows");
  
  Set<int> coords;   // set of columns to project to
  
  if (indices.empty()) {
    if (ambient_dim==dim)
      throw std::runtime_error("p_proj: polyhedron full-dimensional, don't know where to project on");
    bool found=false;
    for (Entire< Subsets_of_k <sequence> >::const_iterator i=entire(all_subsets_of_k(sequence(1,ambient_dim),codim));!found&&!i.at_end(); ++i)
      if (det(affine_hull.minor(All,*i))!=0) {
	coords=*i;
	found=true;
      }
    if (!found) throw std::runtime_error("p_proj: no non-singular minor in AFFINE_HULL");
  } else {
    for (Entire< Array<int> >::const_iterator i=entire(indices); !i.at_end(); ++i) {
      if (*i<1 || *i>ambient_dim)
	throw std::runtime_error("p_proj: index out of range");
      coords+=*i;
    }
    if (!options["revert"])
      coords=range(1,ambient_dim)-coords;
  }
  
  Matrix<Rational> Points, Inequalities;
  bool points_read=false, inequalities_read=false;
  
  perl::Object p_out("Polytope<Rational>");

  if (p_in.lookup("VERTICES | POINTS") >> Points) {
    points_read=true;
    p_out.take("POINTS") << Points.minor(All,~coords);
  }
  
  if (!points_read && options["nofm"])
    throw std::runtime_error("p_proj: no points found and Fourier-Motzkin elimination excluded");
  
  if (!options["nofm"]) {
    if (p_in.lookup("FACETS | INEQUALITIES") >> Inequalities) {
      inequalities_read=true;
      Inequalities /= affine_hull / (-affine_hull);
    }
    
    if (inequalities_read) {
       // perform Fourier-Motzkin elimination on coords
      for (Entire< Set<int> >::const_reverse_iterator c=rentire(coords); !c.at_end(); ++c) {
	Set<int> negative, zero, positive;
	// normalize Inequalities such that there is a -1,0, or 1 in the next column to be eliminated
	for (int i=0; i<Inequalities.rows(); ++i) {
	  const Rational x=Inequalities(i,*c);
	  const int s=sign(x);
	  if (s) {
	    if (s<0)
		     negative += i;
	    else
	      positive += i;
	    Inequalities[i] /= abs(x);
	  } else
	    zero += i;
	}
	const Set<int> remaining_coords=range(0,Inequalities.cols()-1)-(*c);
	Inequalities = Inequalities.minor(All,remaining_coords);
	ListMatrix< Vector<Rational> > Combined_Ineqs=Inequalities.minor(zero,All);
	for (Entire< Set<int> >::const_iterator i=entire(negative); !i.at_end(); ++i)
	  for (Entire< Set<int> >::const_iterator k=entire(positive); !k.at_end(); ++k)
	    Combined_Ineqs /= Inequalities[*i]+Inequalities[*k];
	// delete redundant rows
	for (Entire< Rows< ListMatrix< Vector<Rational> > > >::iterator i=entire(rows(Combined_Ineqs)); !i.at_end(); ) {
	  const Rows< ListMatrix< Vector<Rational> > >::iterator j(i++);
	  if (!*j) Combined_Ineqs.delete_row(j);
	}
	Inequalities = Combined_Ineqs;
      }
      p_out.take("INEQUALITIES") << Inequalities;
    }
  }
  return p_out;
}

perl::Object projection_full(perl::Object p, perl::OptionSet options)
{
  const Array<int> empty;
  return projection(p,empty,options);
}


UserFunction4perl("# Category: Producing a new polyhedron from others\n"
		  "# Orthogonally project a polyhedron to a coordinate subspace.\n\n"
		  "# The subspace the polyhedron is projected on is given by the coordinate indices in the set @a Indices.\n"
		  "# The option @c revert inverts the coordinate list. If no coordinates are specified the polytope is\n"
		  "# projected by omitting \"redundant\" columns, i.e., the projection becomes full-dimensional without\n"
		  "# changing the combinatorial type.\n\n"
		  "# The client scans for all coordinate sections and produces proper output from each.\n"
		  "# If a description in terms of inequalities is found, the client performs Fourier-Motzkin elimination\n"
		  "# unless the @c -nofm option is set.  Setting the @c -nofm option is useful if the corank of the projection\n"
		  "# is large; in this case the number of inequalities produced grows quickly.\n"
		  "# args: Indices [ revert => Bool] [ nofm => Bool]\n",
		  &projection,"projection(Polytope; Array<Int> {revert => 0, nofm => 0})");

UserFunction4perl("# Category: Producing a new polyhedron from others\n"
		  "# Orthogonally project a polyhedron to a coordinate subspace.\n\n"
		  "# The polytope is projected by omitting \"redundant\" columns, i.e., the projection becomes\n"
		  "# full-dimensional without changing the combinatorial type.\n\n"
		  "# The client scans for all coordinate sections and produces proper output from each.\n"
		  "# If a description in terms of inequalities is found, the client performs Fourier-Motzkin elimination\n"
		  "# unless the @c -nofm option is set.  Setting the @c -nofm option is useful if the corank of the projection\n"
		  "# is large; in this case the number of inequalities produced grows quickly.\n"
		  "# args: [ nofm => Bool]\n",
		  &projection_full,"projection_full(Polytope;{revert => 0, nofm => 0})");
} }
