/* Copyright (c) 1997-2008
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ident "$Project: polymake $$Id: wedge.cc 8903 2008-11-10 15:36:39Z gawrilow $"

#include <polymake.h>
#include <Rational.h>
#include <Matrix.h>
#include <linalg.h>
#include <IncidenceMatrix.h>

namespace polymake { namespace polytope {

namespace {

   template <typename PointsSubset, typename Hyperplane> inline
   void lift_to_hyperplane(PointsSubset P, const GenericVector<Hyperplane>& U) {
      typedef typename PointsSubset::value_type point;
      const int d=U.dim()-1;
      for (typename Entire<PointsSubset>::iterator p=entire(P); !p.at_end(); ++p) {
	 typename point::iterator p_i=p->begin();
	 typename Hyperplane::const_iterator u_i=U.top().begin();
	 typename Hyperplane::element_type x(0);
	 for (int i=d; i>0; --i, ++p_i, ++u_i)
	    x -= (*p_i) * (*u_i);
	 (*p_i) = x / (*u_i);
      }
   }

   Matrix<Rational>
   wedge_coord(const Matrix<Rational>& V,		// input vertices
	       const Vector<Rational>& z0,		// interior point
	       const IncidenceMatrix<>& VIF,	// VERTICES_IN_FACETS
	       int edge_facet,
	       const Rational& z,
	       const Rational& z_prime)
   {
      const int d=V.cols(),
	        n_vertices=VIF.cols(),
	        n_vertices_edge=VIF[edge_facet].size(),
	        n_vertices_out= 2*n_vertices-n_vertices_edge;

      Matrix<Rational> V_out=(V / V.minor(~VIF[edge_facet],All)) | zero_vector<Rational>(n_vertices_out);

      // top (or bottom) facet goes thru the barycenter projection and the edge facet
      const Matrix<Rational> facet=
	 null_space( ( z0                          | (z_prime ? z_prime : z)               ) /
		     (V.minor(VIF[edge_facet],All) | zero_vector<Rational>(n_vertices_edge)) );

      if (z_prime) {
	 lift_to_hyperplane(select(rows(V_out), range(n_vertices,n_vertices_out-1)), facet[0]);

	 if (z)	// bottom side facet has to be inclined too
	    V_out.col(d).slice(sequence(0,n_vertices)-VIF[edge_facet]) =
	       z/z_prime * V_out.col(d).slice(range(n_vertices,n_vertices_out-1));
      } else {
	 lift_to_hyperplane(select(rows(V_out), sequence(0,n_vertices)-VIF[edge_facet]), facet[0]);
      }

      return V_out;
   }

} // end unnamed namespace

   perl::Object wedge(perl::Object p_in, const int edge_facet, const Rational& z, const Rational& z2, perl::OptionSet options)
   {
      const Rational z_prime= z2==0 && z==0 ? Rational(1) : z2;

      if (!options["noc"] && z==z_prime)
	 throw std::runtime_error("wedge: z and z' must not be equal");

      const IncidenceMatrix<> VIF=p_in.give("VERTICES_IN_FACETS");
      const int n_vertices=VIF.cols(), n_facets=VIF.rows();
  
      if (edge_facet < 0 || edge_facet >= n_facets)
	 throw std::runtime_error("wedge: edge facet number out of range");
  
      const int n_vertices_edge=VIF[edge_facet].size(),
	        n_vertices_out= 2*n_vertices-n_vertices_edge;

      IncidenceMatrix<> VIF_out=
	 (VIF | VIF.minor(All, ~VIF[edge_facet])) // side facet = original facet + its clone - intersection with the edge
	 / range(0,n_vertices-1);		  // bottom facet = original polytope

      // top facet = edge facet + all clones
      VIF_out[edge_facet]+=range(n_vertices,n_vertices_out-1);

      perl::Object p_out("Polytope<Rational>");
      p_out.set_description() << "wedge over " << p_in.name() << "; edge facet " << edge_facet << endl;

      if (options["noc"]) {
	 if (p_in.exists("DIM")) {
	    const int dim=p_in.give("DIM");
	    p_out.take("DIM") << dim+1;
	 }
      }

      p_out.take("N_VERTICES") << VIF_out.cols(); 
      p_out.take("VERTICES_IN_FACETS") << VIF_out;

      if (options["relabel"]) {
	 std::vector<std::string> labels(n_vertices_out);
	 read_labels(p_in, "VERTEX_LABELS", labels);
	 const std::string tick="'";
	 std::copy(entire(attach_operation(select(labels, sequence(0,n_vertices)-VIF[edge_facet]),
					   constant(tick), operations::add())),
		   labels.begin()+n_vertices);
	 p_out.take("VERTEX_LABELS") << labels;
      }

      if (!options["no_coordinates"]) {
	 const Matrix<Rational> V=p_in.give("VERTICES");
	 const Vector<Rational> z0=p_in.give("VERTEX_BARYCENTER");
	 const Matrix<Rational> V_out=wedge_coord(V,z0,VIF,edge_facet,z,z_prime);
	 p_out.take("VERTICES") << V_out;
	 p_out.take("BOUNDED") << true;
      }
      return p_out;    
   }

   UserFunction4perl("# Category: Producing a new polyhedron from others\n"
		     "# Make a wedge from a polytope over its facet.\n"
		     "#  $facet$ is the number of the `cutting edge' facet. The inclination of the bottom and top side facet\n"
		     "# is controlled by $z$ and $z\'$ paremeters, where $z$ is the height of the projection of the old vertex\n"
		     "#  barycenter on the bottom side facet, and $z\'$ - on the top one.\n"
		     "# Default heights are $[0,1]$; if only $z$ is specified and non-zero, then $z'$ defaults to $-z$.\n"
		     "#  The polytope must be bounded.\n"
		     "#  If the option @c noc is set, then only combinatorial information is handled.\n"
		     "#  The option @c relabel creates an additional section @see VERTEX_LABELS.\n"
		     "#  The bottom facet vertices get the labels from the original polytope;\n"
		     "#  the labels of their clones in the top facet get a tick (') appended.\n"
		     "# args: facet [z [z\'] ] [ noc => Bool, relabel => Bool} ]\n"
		     "# @author Kerstin Fritzsche (initial version)\n",
		     &wedge,"wedge(Polytope, $; $=0, $=1, { noc => undef, relabel => undef})");
} }
