/* Copyright (c) 1997-2008
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ifndef _POLYMAKE_PERL_SCALAR_H
#define _POLYMAKE_PERL_SCALAR_H "$Project: polymake $$Id: Scalar.h 8414 2008-09-04 15:21:25Z gawrilow $"

#include <string>
#include <assoc.h>
#include <perl/glue.h>
#include <typeinfo>

namespace pm { namespace perl {

   class ScalarBase {
   protected:
      SV *sv;

      ScalarBase() : sv(pm_perl_newSV()) {}

      explicit ScalarBase(SV *sv_arg)
	 : sv(sv_arg) {}

      void forget() { pm_perl_decr_SV(sv); }
   public:
      SV* get() const { return sv; }
      SV* get_temp() { return pm_perl_2mortal(sv); }
   };

   class Scalar : public ScalarBase {
   public:
      Scalar() {}
      Scalar(const Scalar& x) { pm_perl_set_copy(sv,x.sv); }
      Scalar& operator= (const Scalar& x) { pm_perl_set_copy(sv,x.sv); return *this; }
      ~Scalar() { forget(); }
   };

   class Value;

   class ArrayBase : public ScalarBase {
   public:
      explicit ArrayBase(int prealloc=0)
	 : ScalarBase(pm_perl_newAV(prealloc)) {}

      explicit ArrayBase(SV *sv_arg)
	 : ScalarBase(sv_arg) {}

      explicit ArrayBase(const Value&);

      ArrayBase(const ScalarBase& x, int prealloc)
	 : ScalarBase(x)
      {
	 pm_perl_makeAV(sv,prealloc);
      }

      ArrayBase(SV *sv_arg, value_flags flags)
	 : ScalarBase(sv_arg)
      {
	 if (flags & value_not_trusted) verify();
      }

      void upgrade(int prealloc)
      {
	 pm_perl_makeAV(sv,prealloc);
      }

      void verify() const
      {
	 if (!pm_perl_is_AV_reference(sv))
	    throw std::runtime_error("input argument is not an array");
      }

      int size() const { return pm_perl_AV_size(sv); }

      SV* operator[] (int i) const
      {
	 return *pm_perl_AV_fetch(sv,i);
      }

      SV* shift()
      {
	 return pm_perl_AV_shift(sv);
      }

      SV* pop()
      {
	 return pm_perl_AV_pop(sv);
      }

      void unshift(const ScalarBase& x)
      {
	 pm_perl_AV_unshift(sv,x.get());
      }

      void push(const ScalarBase& x)
      {
	 pm_perl_AV_push(sv,x.get());
      }
   };

   class HashBase : public ScalarBase {
   protected:
      HashBase() {}
   public:
      explicit HashBase(SV *sv_arg)
	 : ScalarBase(sv_arg) {}

      void verify()
      {
	 if (!pm_perl_is_HV_reference(sv))
	    throw std::runtime_error("input argument is not a hash");
      }

      bool exists(const std::string& key) const
      {
	 return pm_perl_HV_exists(sv, key.c_str(), key.size());
      }

      template <size_t klen>
      bool exists (const char (&key)[klen]) const
      {
	 return pm_perl_HV_exists(sv, key, klen-1);
      }
   };
} }

#endif // _POLYMAKE_PERL_SCALAR_H

// Local Variables:
// mode:C++
// End:
